import base64
import dataclasses
import re
from enum import Enum
from typing import Any, NewType, TypeVar

import pytest
from pydantic import BaseModel, Field, ValidationError

import strawberry
from strawberry.experimental.pydantic._compat import (
    IS_PYDANTIC_V2,
    CompatModelField,
    PydanticCompat,
)
from strawberry.experimental.pydantic.exceptions import (
    AutoFieldsNotInBaseModelError,
    BothDefaultAndDefaultFactoryDefinedError,
)
from strawberry.experimental.pydantic.utils import get_default_factory_for_field
from strawberry.types.base import (
    StrawberryList,
    StrawberryObjectDefinition,
    StrawberryOptional,
)
from tests.experimental.pydantic.utils import needs_pydantic_v1

if IS_PYDANTIC_V2:
    from pydantic import computed_field


def test_can_use_type_standalone():
    class User(BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: strawberry.auto

    user = UserType(age=1, password="abc")

    assert user.age == 1
    assert user.password == "abc"


def test_can_convert_pydantic_type_to_strawberry():
    class User(BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: strawberry.auto

    origin_user = User(age=1, password="abc")
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.password == "abc"


def test_cannot_convert_pydantic_type_to_strawberry_missing_field():
    class User(BaseModel):
        age: int

    with pytest.raises(
        AutoFieldsNotInBaseModelError,
        match=re.escape(
            "UserType defines ['password'] with strawberry.auto."
            " Field(s) not present in User BaseModel."
        ),
    ):

        @strawberry.experimental.pydantic.type(User)
        class UserType:
            age: strawberry.auto
            password: strawberry.auto


def test_cannot_convert_pydantic_type_to_strawberry_property_auto():
    # auto inferring type of a property is not supported

    class User(BaseModel):
        age: int

        @property
        def password(self) -> str:
            return "hunter2"

    with pytest.raises(
        AutoFieldsNotInBaseModelError,
        match=re.escape(
            "UserType defines ['password'] with strawberry.auto."
            " Field(s) not present in User BaseModel."
        ),
    ):

        @strawberry.experimental.pydantic.type(User)
        class UserType:
            age: strawberry.auto
            password: strawberry.auto


def test_can_convert_pydantic_type_to_strawberry_property():
    class User(BaseModel):
        age: int

        @property
        def password(self) -> str:
            return "hunter2"

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: str

    origin_user = User(age=1)
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.password == "hunter2"


def test_can_convert_alias_pydantic_field_to_strawberry():
    class UserModel(BaseModel):
        age_: int = Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age_: strawberry.auto
        password: strawberry.auto

    origin_user = UserModel(age=1, password="abc")
    user = User.from_pydantic(origin_user)

    assert user.age_ == 1
    assert user.password == "abc"


def test_convert_alias_name():
    class UserModel(BaseModel):
        age_: int = Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(
        UserModel, all_fields=True, use_pydantic_alias=True
    )
    class User: ...

    origin_user = UserModel(age=1, password="abc")
    user = User.from_pydantic(origin_user)
    assert user.age_ == 1
    definition = User.__strawberry_definition__

    assert definition.fields[0].graphql_name == "age"


def test_do_not_convert_alias_name():
    class UserModel(BaseModel):
        age_: int = Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(
        UserModel, all_fields=True, use_pydantic_alias=False
    )
    class User: ...

    origin_user = UserModel(age=1, password="abc")
    user = User.from_pydantic(origin_user)
    assert user.age_ == 1
    definition = User.__strawberry_definition__

    assert definition.fields[0].graphql_name is None


def test_can_pass_pydantic_field_description_to_strawberry():
    class UserModel(BaseModel):
        age: int
        password: str | None = Field(..., description="NOT 'password'.")

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.auto

    definition = User.__strawberry_definition__

    assert definition.fields[0].python_name == "age"
    assert definition.fields[0].description is None

    assert definition.fields[1].python_name == "password"
    assert definition.fields[1].description == "NOT 'password'."


def test_can_convert_falsy_values_to_strawberry():
    class UserModel(BaseModel):
        age: int
        password: str

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.auto

    origin_user = UserModel(age=0, password="")
    user = User.from_pydantic(origin_user)

    assert user.age == 0
    assert not user.password


def test_can_convert_pydantic_type_to_strawberry_with_private_field():
    class UserModel(BaseModel):
        age: int

    @strawberry.experimental.pydantic.type(model=UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.Private[str]

    user = User(age=30, password="qwerty")
    assert user.age == 30
    assert user.password == "qwerty"

    definition = User.__strawberry_definition__
    assert len(definition.fields) == 1
    assert definition.fields[0].python_name == "age"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is int


def test_can_convert_pydantic_type_with_nested_data_to_strawberry():
    class WorkModel(BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        name: strawberry.auto

    class UserModel(BaseModel):
        work: WorkModel

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(work=WorkModel(name="Ice Cream inc"))
    user = User.from_pydantic(origin_user)

    assert user.work.name == "Ice Cream inc"


def test_can_convert_pydantic_type_with_list_of_nested_data_to_strawberry():
    class WorkModel(BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        name: strawberry.auto

    class UserModel(BaseModel):
        work: list[WorkModel]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(
        work=[
            WorkModel(name="Ice Cream inc"),
            WorkModel(name="Wall Street"),
        ]
    )
    user = User.from_pydantic(origin_user)

    assert user.work == [Work(name="Ice Cream inc"), Work(name="Wall Street")]


def test_can_convert_pydantic_type_with_list_of_nested_int_to_strawberry():
    class UserModel(BaseModel):
        hours: list[int]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        hours: strawberry.auto

    origin_user = UserModel(
        hours=[
            8,
            9,
            10,
        ]
    )
    user = User.from_pydantic(origin_user)

    assert user.hours == [8, 9, 10]


def test_can_convert_pydantic_type_with_matrix_list_of_nested_int_to_strawberry():
    class UserModel(BaseModel):
        hours: list[list[int]]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        hours: strawberry.auto

    origin_user = UserModel(
        hours=[
            [8, 10],
            [9, 11],
            [10, 12],
        ]
    )
    user = User.from_pydantic(origin_user)

    assert user.hours == [
        [8, 10],
        [9, 11],
        [10, 12],
    ]


def test_can_convert_pydantic_type_with_matrix_list_of_nested_model_to_strawberry():
    class HourModel(BaseModel):
        hour: int

    @strawberry.experimental.pydantic.type(HourModel)
    class Hour:
        hour: strawberry.auto

    class UserModel(BaseModel):
        hours: list[list[HourModel]]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        hours: strawberry.auto

    origin_user = UserModel(
        hours=[
            [
                HourModel(hour=1),
                HourModel(hour=2),
            ],
            [
                HourModel(hour=3),
                HourModel(hour=4),
            ],
            [
                HourModel(hour=5),
                HourModel(hour=6),
            ],
        ]
    )
    user = User.from_pydantic(origin_user)

    assert user.hours == [
        [
            Hour(hour=1),
            Hour(hour=2),
        ],
        [
            Hour(hour=3),
            Hour(hour=4),
        ],
        [
            Hour(hour=5),
            Hour(hour=6),
        ],
    ]


def test_can_convert_pydantic_type_to_strawberry_with_union():
    class BranchA(BaseModel):
        field_a: str

    class BranchB(BaseModel):
        field_b: int

    class User(BaseModel):
        age: int
        union_field: BranchA | BranchB

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType:
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType:
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        union_field: strawberry.auto

    origin_user = User(age=1, union_field=BranchA(field_a="abc"))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchAType)
    assert user.union_field.field_a == "abc"

    origin_user = User(age=1, union_field=BranchB(field_b=123))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchBType)
    assert user.union_field.field_b == 123


def test_can_convert_pydantic_type_to_strawberry_with_union_of_strawberry_types():
    @strawberry.type
    class BranchA:
        field_a: str

    @strawberry.type
    class BranchB:
        field_b: int

    class User(BaseModel):
        age: int
        union_field: BranchA | BranchB

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        union_field: strawberry.auto

    origin_user = User(age=1, union_field=BranchA(field_a="abc"))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchA)
    assert user.union_field.field_a == "abc"

    origin_user = User(age=1, union_field=BranchB(field_b=123))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchB)
    assert user.union_field.field_b == 123


def test_can_convert_pydantic_type_to_strawberry_with_union_nullable():
    class BranchA(BaseModel):
        field_a: str

    class BranchB(BaseModel):
        field_b: int

    class User(BaseModel):
        age: int
        union_field: None | BranchA | BranchB

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType:
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType:
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        union_field: strawberry.auto

    origin_user = User(age=1, union_field=BranchA(field_a="abc"))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchAType)
    assert user.union_field.field_a == "abc"

    origin_user = User(age=1, union_field=BranchB(field_b=123))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.union_field, BranchBType)
    assert user.union_field.field_b == 123

    origin_user = User(age=1, union_field=None)
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.union_field is None


def test_can_convert_pydantic_type_to_strawberry_with_enum():
    @strawberry.enum
    class UserKind(Enum):
        user = 0
        admin = 1

    class User(BaseModel):
        age: int
        kind: UserKind

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        kind: strawberry.auto

    origin_user = User(age=1, kind=UserKind.user)
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.kind == UserKind.user


def test_can_convert_pydantic_type_to_strawberry_with_interface():
    class Base(BaseModel):
        base_field: str

    class BranchA(Base):
        field_a: str

    class BranchB(Base):
        field_b: int

    class User(BaseModel):
        age: int
        interface_field: Base

    @strawberry.experimental.pydantic.interface(Base)
    class BaseType:
        base_field: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType(BaseType):
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType(BaseType):
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        interface_field: strawberry.auto

    origin_user = User(age=1, interface_field=BranchA(field_a="abc", base_field="def"))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.interface_field, BranchAType)
    assert user.interface_field.field_a == "abc"

    origin_user = User(age=1, interface_field=BranchB(field_b=123, base_field="def"))
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert isinstance(user.interface_field, BranchBType)
    assert user.interface_field.field_b == 123


def test_can_convert_pydantic_type_to_strawberry_with_additional_fields():
    class UserModel(BaseModel):
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: int
        password: strawberry.auto

    origin_user = UserModel(password="abc")
    user = User.from_pydantic(origin_user, extra={"age": 1})

    assert user.age == 1
    assert user.password == "abc"


def test_can_convert_pydantic_type_to_strawberry_with_additional_nested_fields():
    @strawberry.type
    class Work:
        name: str

    class UserModel(BaseModel):
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: Work
        password: strawberry.auto

    origin_user = UserModel(password="abc")
    user = User.from_pydantic(origin_user, extra={"work": {"name": "Ice inc"}})

    assert user.work.name == "Ice inc"
    assert user.password == "abc"


def test_can_convert_pydantic_type_to_strawberry_with_additional_list_nested_fields():
    @strawberry.type
    class Work:
        name: str

    class UserModel(BaseModel):
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: list[Work]
        password: strawberry.auto

    origin_user = UserModel(password="abc")
    user = User.from_pydantic(
        origin_user,
        extra={
            "work": [
                {"name": "Software inc"},
                {"name": "Homemade inc"},
            ]
        },
    )

    assert user.work == [
        Work(name="Software inc"),
        Work(name="Homemade inc"),
    ]
    assert user.password == "abc"


def test_can_convert_pydantic_type_to_strawberry_with_missing_data_in_nested_type():
    class WorkModel(BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        year: int
        name: strawberry.auto

    class UserModel(BaseModel):
        work: list[WorkModel]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(work=[WorkModel(name="Software inc")])

    user = User.from_pydantic(
        origin_user,
        extra={
            "work": [
                {"year": 2020},
            ]
        },
    )

    assert user.work == [
        Work(name="Software inc", year=2020),
    ]


def test_can_convert_pydantic_type_to_strawberry_with_missing_index_data_nested_type():
    class WorkModel(BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        year: int
        name: strawberry.auto

    class UserModel(BaseModel):
        work: list[WorkModel | None]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(
        work=[
            WorkModel(name="Software inc"),
            None,
        ]
    )

    user = User.from_pydantic(
        origin_user,
        extra={
            "work": [
                {"year": 2020},
                {"name": "Alternative", "year": 3030},
            ]
        },
    )

    assert user.work == [
        Work(name="Software inc", year=2020),
        # This was None in the UserModel
        Work(name="Alternative", year=3030),
    ]


def test_can_convert_pydantic_type_to_strawberry_with_optional_list():
    class WorkModel(BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        name: strawberry.auto
        year: int

    class UserModel(BaseModel):
        work: WorkModel | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(work=None)

    user = User.from_pydantic(
        origin_user,
    )

    assert user.work is None


def test_can_convert_pydantic_type_to_strawberry_with_optional_nested_value():
    class UserModel(BaseModel):
        names: list[str] | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        names: strawberry.auto

    origin_user = UserModel(names=None)

    user = User.from_pydantic(
        origin_user,
    )

    assert user.names is None


def test_can_convert_input_types_to_pydantic():
    class User(BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        age: strawberry.auto
        password: strawberry.auto

    data = UserInput(age=1, password=None)
    user = data.to_pydantic()

    assert user.age == 1
    assert user.password is None


def test_can_convert_input_types_to_pydantic_default_values():
    class User(BaseModel):
        age: int
        password: str | None = None

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        age: strawberry.auto
        password: strawberry.auto

    data = UserInput(age=1)
    user = data.to_pydantic()

    assert user.age == 1
    assert user.password is None


def test_can_convert_input_types_to_pydantic_default_values_defaults_declared_first():
    # test that we can declare a field with a default. before a field without a default
    class User(BaseModel):
        password: str | None = None
        age: int

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        password: strawberry.auto
        age: strawberry.auto

    data = UserInput(age=1)
    user = data.to_pydantic()

    assert user.age == 1
    assert user.password is None

    definition: StrawberryObjectDefinition = UserInput.__strawberry_definition__
    assert definition.name == "UserInput"

    [
        password_field,
        age_field,
    ] = definition.fields

    assert age_field.python_name == "age"
    assert age_field.type is int

    assert password_field.python_name == "password"
    assert isinstance(password_field.type, StrawberryOptional)
    assert password_field.type.of_type is str


def test_can_convert_pydantic_type_to_strawberry_newtype():
    Password = NewType("Password", str)

    class User(BaseModel):
        age: int
        password: Password | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: strawberry.auto

    origin_user = User(age=1, password="abc")
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.password == "abc"


def test_can_convert_pydantic_type_to_strawberry_newtype_list():
    Password = NewType("Password", str)

    class User(BaseModel):
        age: int
        passwords: list[Password]

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        passwords: strawberry.auto

    origin_user = User(age=1, passwords=["hunter2"])
    user = UserType.from_pydantic(origin_user)

    assert user.age == 1
    assert user.passwords == ["hunter2"]


def test_get_default_factory_for_field():
    class User(BaseModel):
        pass

    compat = PydanticCompat.from_model(User)
    MISSING_TYPE = compat.PYDANTIC_MISSING_TYPE

    def _get_field(
        default: Any = MISSING_TYPE,
        default_factory: Any = MISSING_TYPE,
    ) -> CompatModelField:
        return CompatModelField(
            name="a",
            type_=str,
            outer_type_=str,
            default=default,
            default_factory=default_factory,
            alias="a",
            allow_none=False,
            description="",
            has_alias=False,
            required=True,
            _missing_type=MISSING_TYPE,
            is_v1=not IS_PYDANTIC_V2,
        )

    field = _get_field()

    assert get_default_factory_for_field(field, compat) is dataclasses.MISSING

    def factory_func():
        return "strawberry"

    field = _get_field(default_factory=factory_func)

    # should return the default_factory unchanged
    assert get_default_factory_for_field(field, compat) is factory_func

    mutable_default = [123, "strawberry"]

    field = _get_field(mutable_default)

    created_factory = get_default_factory_for_field(field, compat)

    # should return a factory that copies the default parameter
    assert created_factory() == mutable_default
    assert created_factory() is not mutable_default

    field = _get_field(default=mutable_default, default_factory=factory_func)

    with pytest.raises(
        BothDefaultAndDefaultFactoryDefinedError,
        match=r"Not allowed to specify both default and default_factory.",
    ):
        get_default_factory_for_field(field, compat)


def test_convert_input_types_to_pydantic_default_and_default_factory():
    # Pydantic should raise an error if the user specifies both default
    # and default_factory. this checks for a regression on their side
    if IS_PYDANTIC_V2:
        with pytest.raises(
            TypeError,
            match=("cannot specify both default and default_factory"),
        ):

            class User(BaseModel):
                password: str | None = Field(default=None, default_factory=lambda: None)

    else:
        with pytest.raises(
            ValueError,
            match=("cannot specify both default and default_factory"),
        ):

            class User(BaseModel):
                password: str | None = Field(default=None, default_factory=lambda: None)


def test_can_convert_pydantic_type_to_strawberry_with_additional_field_resolvers():
    def some_resolver() -> int:
        return 84

    class UserModel(BaseModel):
        password: str | None
        new_age: int

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        password: strawberry.auto
        new_age: int = strawberry.field(resolver=some_resolver)

        @strawberry.field
        def age() -> int:
            return 42

    origin_user = UserModel(password="abc", new_age=21)
    user = User.from_pydantic(origin_user)
    assert user.password == "abc"
    assert User.__strawberry_definition__.fields[0].name == "age"
    assert User.__strawberry_definition__.fields[0].base_resolver() == 42
    assert User.__strawberry_definition__.fields[2].name == "new_age"
    assert User.__strawberry_definition__.fields[2].base_resolver() == 84


def test_can_convert_both_output_and_input_type():
    class Work(BaseModel):
        time: float

    class User(BaseModel):
        name: str
        work: Work | None

    class Group(BaseModel):
        users: list[User]

    # Test both definition orders
    @strawberry.experimental.pydantic.input(Work)
    class WorkInput:
        time: strawberry.auto

    @strawberry.experimental.pydantic.type(Work)
    class WorkOutput:
        time: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserOutput:
        name: strawberry.auto
        work: strawberry.auto

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        name: strawberry.auto
        work: strawberry.auto

    @strawberry.experimental.pydantic.input(Group)
    class GroupInput:
        users: strawberry.auto

    @strawberry.experimental.pydantic.type(Group)
    class GroupOutput:
        users: strawberry.auto

    origin_group = Group(
        users=[
            User(name="Alice", work=Work(time=10.0)),
            User(name="Bob", work=Work(time=5.0)),
        ]
    )
    group = GroupOutput.from_pydantic(origin_group)
    final_group = group.to_pydantic()
    assert origin_group == final_group

    group_input = GroupInput.from_pydantic(origin_group)
    final_group = group_input.to_pydantic()
    assert origin_group == final_group


def test_custom_conversion_functions():
    class User(BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: str
        password: strawberry.auto

        @staticmethod
        def from_pydantic(
            instance: User, extra: dict[str, Any] | None = None
        ) -> "UserType":
            return UserType(
                age=str(instance.age),
                password=base64.b64encode(instance.password.encode()).decode()
                if instance.password
                else None,
            )

        def to_pydantic(self) -> User:
            return User(
                age=int(self.age),
                password=base64.b64decode(self.password.encode()).decode()
                if self.password
                else None,
            )

    user = User(age=1, password="abc")
    user_strawberry = UserType.from_pydantic(user)

    assert user_strawberry.age == "1"
    assert user_strawberry.password == "YWJj"

    user_pydantic = user_strawberry.to_pydantic()
    assert user == user_pydantic


def test_nested_custom_conversion_functions():
    class User(BaseModel):
        age: int
        password: str | None

    class Parent(BaseModel):
        user: User

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: str
        password: strawberry.auto

        @staticmethod
        def from_pydantic(
            instance: User, extra: dict[str, Any] | None = None
        ) -> "UserType":
            return UserType(
                age=str(instance.age),
                password=base64.b64encode(instance.password.encode()).decode()
                if instance.password
                else None,
            )

        def to_pydantic(self) -> User:
            return User(
                age=int(self.age),
                password=base64.b64decode(self.password.encode()).decode()
                if self.password
                else None,
            )

    @strawberry.experimental.pydantic.type(Parent)
    class ParentType:
        user: strawberry.auto

    user = User(age=1, password="abc")
    parent = Parent(user=user)
    parent_strawberry = ParentType.from_pydantic(parent)

    assert parent_strawberry.user.age == "1"
    assert parent_strawberry.user.password == "YWJj"

    parent_pydantic = parent_strawberry.to_pydantic()
    assert parent == parent_pydantic


def test_can_convert_input_types_to_pydantic_with_non_pydantic_dataclass():
    @strawberry.type
    class Work:
        hours: int

    class User(BaseModel):
        age: int
        password: str | None
        work: Work

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        age: strawberry.auto
        password: strawberry.auto
        work: strawberry.auto

    data = UserInput(age=1, password=None, work=Work(hours=1))
    user = data.to_pydantic()

    assert user.age == 1
    assert user.password is None
    assert user.work.hours == 1


def test_can_convert_input_types_to_pydantic_with_dict():
    class Work(BaseModel):
        hours: int

    class User(BaseModel):
        age: int
        password: str | None
        work: dict[str, Work]

    @strawberry.experimental.pydantic.input(Work)
    class WorkInput:
        hours: strawberry.auto

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        age: strawberry.auto
        password: strawberry.auto
        work: strawberry.auto

    data = UserInput(age=1, password=None, work={"Monday": Work(hours=1)})
    user = data.to_pydantic()

    assert user.age == 1
    assert user.password is None
    assert user.work["Monday"].hours == 1


def test_can_add_missing_arguments_to_pydantic():
    class User(BaseModel):
        age: int
        password: str

    @strawberry.experimental.pydantic.type(User)
    class UserInput:
        age: strawberry.auto

    data = UserInput(age=1)
    user = data.to_pydantic(password="hunter2")

    assert user.age == 1
    assert user.password == "hunter2"


def test_raise_missing_arguments_to_pydantic():
    class User(BaseModel):
        age: int
        password: str

    @strawberry.experimental.pydantic.type(User)
    class UserInput:
        age: strawberry.auto

    data = UserInput(age=1)

    with pytest.raises(
        ValidationError,
        match=("1 validation error for User"),
    ):
        data.to_pydantic()


def test_can_convert_generic_alias_fields_to_strawberry():
    class TestModel(BaseModel):
        list_1d: list[int]
        list_2d: list[list[int]]

    @strawberry.experimental.pydantic.type(TestModel)
    class Test:
        list_1d: strawberry.auto
        list_2d: strawberry.auto

    fields = Test.__strawberry_definition__.fields
    assert isinstance(fields[0].type, StrawberryList)
    assert isinstance(fields[1].type, StrawberryList)

    model = TestModel(
        list_1d=[1, 2, 3],
        list_2d=[[1, 2], [3]],
    )
    test = Test.from_pydantic(model)

    assert test.list_1d == [1, 2, 3]
    assert test.list_2d == [[1, 2], [3]]


def test_can_convert_optional_union_type_expression_fields_to_strawberry():
    class TestModel(BaseModel):
        optional_list: list[int] | None
        optional_str: str | None

    @strawberry.experimental.pydantic.type(TestModel)
    class Test:
        optional_list: strawberry.auto
        optional_str: strawberry.auto

    fields = Test.__strawberry_definition__.fields
    assert isinstance(fields[0].type, StrawberryOptional)
    assert isinstance(fields[1].type, StrawberryOptional)

    model = TestModel(
        optional_list=[1, 2, 3],
        optional_str=None,
    )
    test = Test.from_pydantic(model)

    assert test.optional_list == [1, 2, 3]
    assert test.optional_str is None


@needs_pydantic_v1
def test_can_convert_pydantic_type_to_strawberry_with_constrained_list():
    from pydantic import ConstrainedList

    class WorkModel(BaseModel):
        name: str

    class WorkList(ConstrainedList):
        min_items = 1

    class UserModel(BaseModel):
        work: WorkList[WorkModel]

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        name: strawberry.auto

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(
        work=[WorkModel(name="developer"), WorkModel(name="tester")]
    )

    user = User.from_pydantic(origin_user)

    assert user == User(work=[Work(name="developer"), Work(name="tester")])


SI = TypeVar("SI", covariant=True)  # pragma: no mutate


class SpecialList(list[SI]):
    pass


@needs_pydantic_v1
def test_can_convert_pydantic_type_to_strawberry_with_specialized_list():
    class WorkModel(BaseModel):
        name: str

    class WorkList(SpecialList[SI]):
        min_items = 1

    class UserModel(BaseModel):
        work: WorkList[WorkModel]

    @strawberry.experimental.pydantic.type(WorkModel)
    class Work:
        name: strawberry.auto

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        work: strawberry.auto

    origin_user = UserModel(
        work=[WorkModel(name="developer"), WorkModel(name="tester")]
    )

    user = User.from_pydantic(origin_user)

    assert user == User(work=[Work(name="developer"), Work(name="tester")])


@pytest.mark.skipif(
    not IS_PYDANTIC_V2, reason="Requires Pydantic v2 for computed_field"
)
def test_can_convert_pydantic_type_to_strawberry_computed_field():
    """Test that computed fields on a pydantic type are not accessed unless queried."""

    class UserModel(BaseModel):
        age: int

        @computed_field
        @property
        def name(self) -> str:
            raise Exception("`name` computed_field should not be accessed")

        @computed_field
        @property
        def location(self) -> str:
            return "earth"

    def get_name(root) -> str:
        return root._original_model.name

    def get_location(root) -> str:
        return root._original_model.location

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: int
        name: str = strawberry.field(resolver=get_name)
        location: str = strawberry.field(resolver=get_location)

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User.from_pydantic(UserModel(age=20))

    schema = strawberry.Schema(query=Query)

    query = "{ user { age location } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"] == {"age": 20, "location": "earth"}



import pytest


@pytest.fixture
def test_client():
    from litestar.testing import TestClient

    from tests.litestar.app import create_app

    app = create_app()
    return TestClient(app)


@pytest.fixture
def test_client_keep_alive():
    from litestar.testing import TestClient

    from tests.litestar.app import create_app

    app = create_app(keep_alive=True, keep_alive_interval=0.1)
    return TestClient(app)



from typing import List

class ListLifeGenericResultListLifeItems1:
    name: str
    age: int

class ListLifeGenericResultListLifeItems2:
    name: str
    age: int

class ListLifeGenericResultListLife:
    items1: list[ListLifeGenericResultListLifeItems1]
    items2: list[ListLifeGenericResultListLifeItems2]

class ListLifeGenericResult:
    list_life: ListLifeGenericResultListLife



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from strawberry.django.test import GraphQLTestClient


@pytest.fixture
def graphql_client() -> GraphQLTestClient:
    from django.test.client import Client

    from strawberry.django.test import GraphQLTestClient

    return GraphQLTestClient(Client())



import contextlib

import pytest
from inline_snapshot import snapshot

from tests.conftest import skip_if_gql_32
from tests.http.clients.base import HttpClient

pytestmark = skip_if_gql_32("GraphQL 3.3.0 is required for incremental execution")


async def test_basic_stream(http_client: HttpClient):
    response = await http_client.query(
        method="get",
        query="""
        query Stream {
            streamableField @stream
        }
        """,
    )

    async with contextlib.aclosing(response.streaming_json()) as stream:
        initial = await stream.__anext__()

        assert initial == snapshot(
            {
                "data": {"streamableField": []},
                "hasNext": True,
                "pending": [{"id": "0", "path": ["streamableField"]}],
                "extensions": None,
            }
        )

        first = await stream.__anext__()

        assert first == snapshot(
            {
                "hasNext": True,
                "extensions": None,
                "incremental": [
                    {
                        "items": ["Hello 0"],
                        "id": "0",
                        "path": ["streamableField"],
                        "label": None,
                    }
                ],
            }
        )

        second = await stream.__anext__()

        assert second == snapshot(
            {
                "hasNext": True,
                "extensions": None,
                "incremental": [
                    {
                        "items": ["Hello 1"],
                        "id": "0",
                        "path": ["streamableField"],
                        "label": None,
                    }
                ],
            }
        )

        third = await stream.__anext__()

        assert third == snapshot(
            {"hasNext": False, "extensions": None, "completed": [{"id": "0"}]}
        )

        with pytest.raises(StopAsyncIteration):
            await stream.__anext__()



from pathlib import Path

import pytest
from typer import Typer
from typer.testing import CliRunner

schema = """
type Query {
    hello: String!
}
"""

expected_output = """
from __future__ import annotations
import strawberry

@strawberry.type
class Query:
    hello: str

schema = strawberry.Schema(query=Query)
""".strip()


@pytest.fixture
def schema_file(tmp_path: Path) -> Path:
    schema_file = tmp_path / "schema.graphql"
    schema_file.write_text(schema)
    return schema_file


def test_schema_codegen(cli_app: Typer, cli_runner: CliRunner, schema_file: Path):
    result = cli_runner.invoke(cli_app, ["schema-codegen", str(schema_file)])

    assert result.exit_code == 0
    assert result.stdout.strip() == expected_output


def test_schema_codegen_to_file(
    cli_app: Typer, cli_runner: CliRunner, schema_file: Path, tmp_path: Path
):
    output_file = tmp_path / "schema.py"

    result = cli_runner.invoke(
        cli_app, ["schema-codegen", str(schema_file), "--output", str(output_file)]
    )

    assert "Code generated at `schema.py`" in result.stdout.strip()
    assert result.exit_code == 0
    assert output_file.read_text().strip() == expected_output


def test_overrides_file_if_exists(
    cli_app: Typer, cli_runner: CliRunner, schema_file: Path, tmp_path: Path
):
    output_file = tmp_path / "schema.py"
    output_file.write_text("old content")

    result = cli_runner.invoke(
        cli_app, ["schema-codegen", str(schema_file), "--output", str(output_file)]
    )

    assert "Code generated at `schema.py`" in result.stdout.strip()
    assert result.exit_code == 0
    assert output_file.read_text().strip() == expected_output



import abc
import contextlib
import json
import logging
from collections.abc import AsyncGenerator, AsyncIterable, Callable, Mapping, Sequence
from dataclasses import dataclass
from datetime import timedelta
from functools import cached_property
from io import BytesIO
from typing import Any, Literal, Optional, Union

from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.schema import Schema
from strawberry.subscriptions.protocols.graphql_transport_ws.handlers import (
    BaseGraphQLTransportWSHandler,
)
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    Message as GraphQLTransportWSMessage,
)
from strawberry.subscriptions.protocols.graphql_ws.handlers import BaseGraphQLWSHandler
from strawberry.subscriptions.protocols.graphql_ws.types import OperationMessage
from strawberry.types import ExecutionResult

logger = logging.getLogger("strawberry.test.http_client")

JSON = Union[dict[str, "JSON"], list["JSON"], str, int, float, bool, None]
ResultOverrideFunction = Optional[Callable[[ExecutionResult], GraphQLHTTPResponse]]


@dataclass
class Response:
    status_code: int
    data: bytes | AsyncIterable[bytes]

    def __init__(
        self,
        status_code: int,
        data: bytes | AsyncIterable[bytes],
        *,
        headers: dict[str, str] | None = None,
    ) -> None:
        self.status_code = status_code
        self.data = data
        self._headers = headers or {}

    @cached_property
    def headers(self) -> Mapping[str, str]:
        return {k.lower(): v for k, v in self._headers.items()}

    @property
    def is_multipart(self) -> bool:
        return self.headers.get("content-type", "").startswith("multipart/mixed")

    @property
    def text(self) -> str:
        assert isinstance(self.data, bytes)
        return self.data.decode()

    @property
    def json(self) -> JSON:
        assert isinstance(self.data, bytes)
        return json.loads(self.data)

    async def streaming_json(self) -> AsyncIterable[JSON]:
        if not self.is_multipart:
            raise ValueError("Streaming not supported")

        def parse_chunk(text: str) -> JSON | None:
            # TODO: better parsing? :)
            with contextlib.suppress(json.JSONDecodeError):
                return json.loads(text)

        if isinstance(self.data, AsyncIterable):
            chunks = self.data

            async for chunk in chunks:
                lines = chunk.decode("utf-8").split("\r\n")

                for text in lines:
                    if data := parse_chunk(text):
                        yield data
        else:
            # TODO: we do this because httpx doesn't support streaming
            # it would be nice to fix httpx instead of doing this,
            # but we might have the same issue in other clients too
            # TODO: better message
            logger.warning("Didn't receive a stream, parsing it sync")

            chunks = self.data.decode("utf-8").split("\r\n")

            for chunk in chunks:
                if data := parse_chunk(chunk):
                    yield data


class HttpClient(abc.ABC):
    @abc.abstractmethod
    def __init__(
        self,
        schema: Schema,
        graphiql: bool | None = None,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        keep_alive: bool = False,
        keep_alive_interval: float = 1,
        subscription_protocols: Sequence[str] = (),
        connection_init_wait_timeout: timedelta = timedelta(minutes=1),
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
        max_subscriptions_per_connection: int | None = 100,
    ): ...

    @abc.abstractmethod
    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response: ...

    @abc.abstractmethod
    async def request(
        self,
        url: str,
        method: Literal["head", "get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
    ) -> Response: ...

    @abc.abstractmethod
    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response: ...

    @abc.abstractmethod
    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response: ...

    async def query(
        self,
        query: str,
        method: Literal["get", "post"] = "post",
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
    ) -> Response:
        return await self._graphql_request(
            method,
            query=query,
            operation_name=operation_name,
            headers=headers,
            variables=variables,
            files=files,
            extensions=extensions,
        )

    def _get_headers(
        self,
        method: Literal["get", "post"],
        headers: dict[str, str] | None,
        files: dict[str, BytesIO] | None,
    ) -> dict[str, str]:
        additional_headers = {}
        headers = headers or {}

        # TODO: fix case sensitivity
        content_type = headers.get("content-type")

        if not content_type and method == "post" and not files:
            content_type = "application/json"

        additional_headers = {"Content-Type": content_type} if content_type else {}

        return {**additional_headers, **headers}

    def _build_body(
        self,
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        method: Literal["get", "post"] = "post",
        extensions: dict[str, Any] | None = None,
    ) -> dict[str, object] | None:
        if query is None:
            assert files is None
            assert variables is None

            return None

        body: dict[str, object] = {"query": query}

        if operation_name is not None:
            body["operationName"] = operation_name

        if variables is not None:
            body["variables"] = variables

        if extensions is not None:
            body["extensions"] = extensions

        if files:
            assert variables is not None

            file_map = self._build_multipart_file_map(variables, files)

            body = {
                "operations": json.dumps(body),
                "map": json.dumps(file_map),
            }

        if method == "get" and variables:
            body["variables"] = json.dumps(variables)

        if method == "get" and extensions:
            body["extensions"] = json.dumps(extensions)

        return body

    @staticmethod
    def _build_multipart_file_map(
        variables: dict[str, object], files: dict[str, BytesIO]
    ) -> dict[str, list[str]]:
        # TODO: remove code duplication

        files_map: dict[str, list[str]] = {}
        for key, values in variables.items():
            if isinstance(values, dict):
                folder_key = next(iter(values.keys()))
                key += f".{folder_key}"  # noqa: PLW2901
                # the list of file is inside the folder keyword
                values = values[folder_key]  # noqa: PLW2901

            # If the variable is an array of files we must number the keys
            if isinstance(values, list):
                # copying `files` as when we map a file we must discard from the dict
                _kwargs = files.copy()
                for index, _ in enumerate(values):
                    k = next(iter(_kwargs.keys()))
                    _kwargs.pop(k)
                    files_map.setdefault(k, [])
                    files_map[k].append(f"variables.{key}.{index}")
            else:
                files_map[key] = [f"variables.{key}"]

        return files_map

    def ws_connect(
        self,
        url: str,
        *,
        protocols: list[str],
    ) -> contextlib.AbstractAsyncContextManager["WebSocketClient"]:
        raise NotImplementedError


@dataclass
class Message:
    type: Any
    data: Any
    extra: str | None = None

    def json(self) -> Any:
        return json.loads(self.data)


class WebSocketClient(abc.ABC):
    def name(self) -> str:
        return ""

    @abc.abstractmethod
    async def send_text(self, payload: str) -> None: ...

    @abc.abstractmethod
    async def send_json(self, payload: Mapping[str, object]) -> None: ...

    @abc.abstractmethod
    async def send_bytes(self, payload: bytes) -> None: ...

    @abc.abstractmethod
    async def receive(self, timeout: float | None = None) -> Message: ...

    @abc.abstractmethod
    async def receive_json(self, timeout: float | None = None) -> Any: ...

    @abc.abstractmethod
    async def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def accepted_subprotocol(self) -> str | None: ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool: ...

    @property
    @abc.abstractmethod
    def close_code(self) -> int: ...

    @property
    @abc.abstractmethod
    def close_reason(self) -> str | None: ...

    async def __aiter__(self) -> AsyncGenerator[Message, None]:
        while not self.closed:
            yield await self.receive()

    async def send_message(self, message: GraphQLTransportWSMessage) -> None:
        await self.send_json(message)

    async def send_legacy_message(self, message: OperationMessage) -> None:
        await self.send_json(message)


class DebuggableGraphQLTransportWSHandler(
    BaseGraphQLTransportWSHandler[dict[str, object], object]
):
    def on_init(self) -> None:
        """This method can be patched by unit tests to get the instance of the
        transport handler when it is initialized.
        """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.original_context = kwargs.get("context", {})
        DebuggableGraphQLTransportWSHandler.on_init(self)

    def get_tasks(self) -> list:
        return [op.task for op in self.operations.values()]

    @property
    def context(self):
        self.original_context["ws"] = self.websocket
        self.original_context["get_tasks"] = self.get_tasks
        self.original_context["connectionInitTimeoutTask"] = (
            self.connection_init_timeout_task
        )
        return self.original_context

    @context.setter
    def context(self, value):
        self.original_context = value


class DebuggableGraphQLWSHandler(BaseGraphQLWSHandler[dict[str, object], object]):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.original_context = kwargs.get("context", {})

    def get_tasks(self) -> list:
        return list(self.tasks.values())

    @property
    def context(self):
        self.original_context["ws"] = self.websocket
        self.original_context["get_tasks"] = self.get_tasks
        self.original_context["connectionInitTimeoutTask"] = None
        return self.original_context

    @context.setter
    def context(self, value):
        self.original_context = value



from typing import List

class OperationNameResultPerson:
    name: str

class OperationNameResultListOfPeople:
    name: str

class OperationNameResult:
    person: OperationNameResultPerson
    list_of_people: list[OperationNameResultListOfPeople]



import textwrap
from typing import Generic, TypeVar

import strawberry
from strawberry.scalars import JSON


def test_support_nested_generics():
    T = TypeVar("T")

    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge(Generic[T]):
        node: T

    @strawberry.type
    class Connection(Generic[T]):
        edge: Edge[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> Connection[User]:
            return Connection(edge=Edge(node=User(name="Patrick")))

    schema = strawberry.Schema(query=Query)

    query = """{
        users {
            __typename
            edge {
                __typename
                node {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "users": {
            "__typename": "UserConnection",
            "edge": {"__typename": "UserEdge", "node": {"name": "Patrick"}},
        }
    }


def test_unions_nested_inside_a_list():
    T = TypeVar("T")

    @strawberry.type
    class JsonBlock:
        data: JSON

    @strawberry.type
    class BlockRowtype(Generic[T]):
        total: int
        items: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def blocks(
            self,
        ) -> list[BlockRowtype[int] | BlockRowtype[str] | JsonBlock]:
            return [
                BlockRowtype(total=3, items=["a", "b", "c"]),
                BlockRowtype(total=1, items=[1, 2, 3, 4]),
                JsonBlock(data=JSON({"a": 1})),
            ]

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """query {
        blocks {
            __typename
            ... on IntBlockRowtype {
                a: items
            }
            ... on StrBlockRowtype {
                b: items
            }
            ... on JsonBlock {
                data
            }
        }
    }"""
    )

    assert not result.errors

    assert result.data == {
        "blocks": [
            {"__typename": "StrBlockRowtype", "b": ["a", "b", "c"]},
            {"__typename": "IntBlockRowtype", "a": [1, 2, 3, 4]},
            {"__typename": "JsonBlock", "data": {"a": 1}},
        ]
    }


def test_unions_nested_inside_a_list_with_no_items():
    T = TypeVar("T")

    @strawberry.type
    class JsonBlock:
        data: JSON

    @strawberry.type
    class BlockRowtype(Generic[T]):
        total: int
        items: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def blocks(
            self,
        ) -> list[BlockRowtype[int] | BlockRowtype[str] | JsonBlock]:
            return [
                BlockRowtype(total=3, items=[]),
                BlockRowtype(total=1, items=[]),
                JsonBlock(data=JSON({"a": 1})),
            ]

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """query {
        blocks {
            __typename
            ... on IntBlockRowtype {
                a: items
            }
            ... on StrBlockRowtype {
                b: items
            }
            ... on JsonBlock {
                data
            }
        }
    }"""
    )

    assert not result.errors

    assert result.data == {
        "blocks": [
            {"__typename": "IntBlockRowtype", "a": []},
            {"__typename": "IntBlockRowtype", "a": []},
            {"__typename": "JsonBlock", "data": {"a": 1}},
        ]
    }


def test_unions_nested_inside_a_list_of_lists():
    T = TypeVar("T")

    @strawberry.type
    class JsonBlock:
        data: JSON

    @strawberry.type
    class BlockRowtype(Generic[T]):
        total: int
        items: list[list[T]]

    @strawberry.type
    class Query:
        @strawberry.field
        def blocks(
            self,
        ) -> list[BlockRowtype[int] | BlockRowtype[str] | JsonBlock]:
            return [
                BlockRowtype(total=3, items=[["a", "b", "c"]]),
                BlockRowtype(total=1, items=[[1, 2, 3, 4]]),
                JsonBlock(data=JSON({"a": 1})),
            ]

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """query {
        blocks {
            __typename
            ... on IntBlockRowtype {
                a: items
            }
            ... on StrBlockRowtype {
                b: items
            }
            ... on JsonBlock {
                data
            }
        }
    }"""
    )

    assert not result.errors

    assert result.data == {
        "blocks": [
            {"__typename": "StrBlockRowtype", "b": [["a", "b", "c"]]},
            {"__typename": "IntBlockRowtype", "a": [[1, 2, 3, 4]]},
            {"__typename": "JsonBlock", "data": {"a": 1}},
        ]
    }


def test_using_generics_with_an_interface():
    T = TypeVar("T")

    @strawberry.interface
    class BlockInterface:
        id: strawberry.ID
        disclaimer: str | None = strawberry.field(default=None)

    @strawberry.type
    class JsonBlock(BlockInterface):
        data: JSON

    @strawberry.type
    class BlockRowtype(BlockInterface, Generic[T]):
        total: int
        items: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def blocks(self) -> list[BlockInterface]:
            return [
                BlockRowtype(id=strawberry.ID("3"), total=3, items=["a", "b", "c"]),
                BlockRowtype(id=strawberry.ID("1"), total=1, items=[1, 2, 3, 4]),
                JsonBlock(id=strawberry.ID("2"), data=JSON({"a": 1})),
            ]

    schema = strawberry.Schema(
        query=Query, types=[BlockRowtype[int], JsonBlock, BlockRowtype[str]]
    )

    expected_schema = textwrap.dedent(
        '''
        interface BlockInterface {
          id: ID!
          disclaimer: String
        }

        type IntBlockRowtype implements BlockInterface {
          id: ID!
          disclaimer: String
          total: Int!
          items: [Int!]!
        }

        """
        The `JSON` scalar type represents JSON values as specified by [ECMA-404](https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf).
        """
        scalar JSON @specifiedBy(url: "https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf")

        type JsonBlock implements BlockInterface {
          id: ID!
          disclaimer: String
          data: JSON!
        }

        type Query {
          blocks: [BlockInterface!]!
        }

        type StrBlockRowtype implements BlockInterface {
          id: ID!
          disclaimer: String
          total: Int!
          items: [String!]!
        }
    '''
    ).strip()

    assert str(schema) == expected_schema

    result = schema.execute_sync(
        """query {
        blocks {
            id
            __typename
            ... on IntBlockRowtype {
                a: items
            }
            ... on StrBlockRowtype {
                b: items
            }
            ... on JsonBlock {
                data
            }
        }
    }"""
    )

    assert not result.errors

    assert result.data == {
        "blocks": [
            {"id": "3", "__typename": "StrBlockRowtype", "b": ["a", "b", "c"]},
            {"id": "1", "__typename": "IntBlockRowtype", "a": [1, 2, 3, 4]},
            {"id": "2", "__typename": "JsonBlock", "data": {"a": 1}},
        ]
    }


def test_supports_generic_in_unions_with_nesting():
    T = TypeVar("T")

    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge(Generic[T]):
        node: T

    @strawberry.type
    class Connection(Generic[T]):
        edge: Edge[T]

    @strawberry.type
    class Fallback:
        node: str

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> Connection[User] | Fallback:
            return Connection(edge=Edge(node=User(name="Patrick")))

    schema = strawberry.Schema(query=Query)

    query = """{
        users {
            __typename
            ... on UserConnection {
                edge {
                    __typename
                    node {
                        name
                    }
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "users": {
            "__typename": "UserConnection",
            "edge": {"__typename": "UserEdge", "node": {"name": "Patrick"}},
        }
    }


def test_does_not_raise_duplicated_type_error():
    T = TypeVar("T")

    @strawberry.type
    class Wrapper(Generic[T]):
        value: T

    @strawberry.type
    class Query:
        a: Wrapper[Wrapper[int]]
        b: Wrapper[Wrapper[int]]

    schema = strawberry.Schema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type IntWrapper {
          value: Int!
        }

        type IntWrapperWrapper {
          value: IntWrapper!
        }

        type Query {
          a: IntWrapperWrapper!
          b: IntWrapperWrapper!
        }
    """
    ).strip()
    assert str(schema) == expected_schema



import pytest

import strawberry
from strawberry.schema.config import StrawberryConfig
from tests.conftest import skip_if_gql_32
from tests.http.clients.base import HttpClient
from tests.views.schema import Mutation, MyExtension, Query, Subscription


@pytest.fixture
def batching_http_client(http_client_class: type[HttpClient]) -> HttpClient:
    return http_client_class(
        schema=strawberry.Schema(
            query=Query,
            mutation=Mutation,
            subscription=Subscription,
            extensions=[MyExtension],
            config=StrawberryConfig(batching_config={"max_operations": 10}),
        )
    )


async def test_batching_works_with_multiple_queries(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {"query": "{ hello }"},
            {"query": "{ hello }"},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {"data": {"hello": "Hello world"}, "extensions": {"example": "example"}},
        {"data": {"hello": "Hello world"}, "extensions": {"example": "example"}},
    ]


async def test_batching_works_with_single_query(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {"query": "{ hello }"},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {"data": {"hello": "Hello world"}, "extensions": {"example": "example"}},
    ]


async def test_variables_can_be_supplied_per_query(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {
                "query": "query InjectedVariables($name: String!) { hello(name: $name) }",
                "variables": {"name": "Alice"},
            },
            {
                "query": "query InjectedVariables($name: String!) { hello(name: $name) }",
                "variables": {"name": "Bob"},
            },
            {"query": "query NoVariables{ hello }"},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {"data": {"hello": "Hello Alice"}, "extensions": {"example": "example"}},
        {"data": {"hello": "Hello Bob"}, "extensions": {"example": "example"}},
        {"data": {"hello": "Hello world"}, "extensions": {"example": "example"}},
    ]


async def test_operations_can_be_selected_per_query(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {
                "query": 'query Op1 { hello(name: "Op1") } query Op2 { hello(name: "Op2") }',
                "operationName": "Op1",
            },
            {
                "query": 'query Op1 { hello(name: "Op1") } query Op2 { hello(name: "Op2") }',
                "operationName": "Op2",
            },
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {"data": {"hello": "Hello Op1"}, "extensions": {"example": "example"}},
        {"data": {"hello": "Hello Op2"}, "extensions": {"example": "example"}},
    ]


@skip_if_gql_32("formatting is different in gql 3.2")
async def test_extensions_are_handled_per_query(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {
                "query": 'query { valueFromExtensions(key: "test") }',
                "extensions": {"test": "op1-value"},
            },
            {
                "query": 'query { valueFromExtensions(key: "test") }',
                "extensions": {"test": "op2-value"},
            },
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {
            "data": {"valueFromExtensions": "op1-value"},
            "extensions": {"example": "example"},
        },
        {
            "data": {"valueFromExtensions": "op2-value"},
            "extensions": {"example": "example"},
        },
    ]


async def test_context_is_shared_between_operations(batching_http_client):
    response = await batching_http_client.post(
        url="/graphql",
        json=[
            {"query": 'mutation { updateContext(key: "test", value: "shared-value") }'},
            {"query": 'query { valueFromContext(key: "test") }'},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 200
    assert response.json == [
        {
            "data": {"updateContext": True},
            "extensions": {"example": "example"},
        },
        {
            "data": {"valueFromContext": "shared-value"},
            "extensions": {"example": "example"},
        },
    ]


async def test_returns_error_when_batching_is_disabled(
    http_client_class: type[HttpClient],
):
    http_client = http_client_class(
        schema=strawberry.Schema(
            query=Query,
            mutation=Mutation,
            subscription=Subscription,
            extensions=[MyExtension],
            config=StrawberryConfig(batching_config=None),
        )
    )

    response = await http_client.post(
        url="/graphql",
        json=[
            {"query": "{ hello }"},
            {"query": "{ hello }"},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 400
    assert "Batching is not enabled" in response.text


async def test_returns_error_when_trying_too_many_operations(
    http_client_class: type[HttpClient],
):
    http_client = http_client_class(
        schema=strawberry.Schema(
            query=Query,
            mutation=Mutation,
            subscription=Subscription,
            extensions=[MyExtension],
            config=StrawberryConfig(batching_config={"max_operations": 2}),
        )
    )

    response = await http_client.post(
        url="/graphql",
        json=[
            {"query": "{ hello }"},
            {"query": "{ hello }"},
            {"query": "{ hello }"},
        ],
        headers={"content-type": "application/json"},
    )

    assert response.status_code == 400
    assert "Too many operations" in response.text



import strawberry
from strawberry.directive import DirectiveLocation, DirectiveValue
from strawberry.extensions import SchemaExtension
from strawberry.extensions.directives import (
    DirectivesExtension,
    DirectivesExtensionSync,
)


@strawberry.type
class Query:
    example: str


@strawberry.directive(locations=[DirectiveLocation.FIELD])
def uppercase(value: DirectiveValue[str]) -> str:
    return value.upper()


class MyExtension(SchemaExtension): ...


def test_returns_empty_list_when_no_custom_directives():
    schema = strawberry.Schema(query=Query)

    assert schema.get_extensions() == []


def test_returns_extension_passed_by_user():
    schema = strawberry.Schema(query=Query, extensions=[MyExtension])

    assert len(schema.get_extensions()) == 1
    assert isinstance(schema.get_extensions()[0], MyExtension)


def test_returns_directives_extension_when_passing_directives():
    schema = strawberry.Schema(query=Query, directives=[uppercase])

    assert len(schema.get_extensions()) == 1
    assert isinstance(schema.get_extensions()[0], DirectivesExtension)


def test_returns_extension_passed_by_user_and_directives_extension():
    schema = strawberry.Schema(
        query=Query, extensions=[MyExtension], directives=[uppercase]
    )
    for ext, ext_cls in zip(
        schema.get_extensions(),
        [MyExtension, DirectivesExtension],
        strict=True,
    ):
        assert isinstance(ext, ext_cls)


def test_returns_directives_extension_when_passing_directives_sync():
    schema = strawberry.Schema(query=Query, directives=[uppercase])

    assert len(schema.get_extensions(sync=True)) == 1
    assert isinstance(schema.get_extensions(sync=True)[0], DirectivesExtensionSync)


def test_returns_extension_passed_by_user_and_directives_extension_sync():
    schema = strawberry.Schema(
        query=Query, extensions=[MyExtension], directives=[uppercase]
    )
    for ext, ext_cls in zip(
        schema.get_extensions(sync=True),
        [MyExtension, DirectivesExtensionSync],
        strict=True,
    ):
        assert isinstance(ext, ext_cls)


def test_no_duplicate_extensions_with_directives():
    """Test to verify that extensions are not duplicated when directives are present.

    This test initially fails with the current implementation but passes
    after fixing the get_extensions method.
    """

    schema = strawberry.Schema(
        query=Query, extensions=[MyExtension], directives=[uppercase]
    )

    extensions = schema.get_extensions()

    # Count how many times our extension appears
    ext_count = sum(1 for e in extensions if isinstance(e, MyExtension))

    # With current implementation this fails as ext_count is 2
    assert ext_count == 1, f"Extension appears {ext_count} times instead of once"


def test_extension_order_preserved():
    """Test to verify that extension order is preserved while removing duplicates."""

    class Extension1(SchemaExtension):
        pass

    class Extension2(SchemaExtension):
        pass

    schema = strawberry.Schema(
        query=Query, extensions=[Extension1, Extension2], directives=[uppercase]
    )

    extensions = schema.get_extensions()
    extension_types = [
        type(ext)
        for ext in extensions
        if not isinstance(ext, (DirectivesExtension, DirectivesExtensionSync))
    ]

    assert extension_types == [Extension1, Extension2], "Extension order not preserved"



from __future__ import annotations

from typing import Annotated
from uuid import UUID

import strawberry


@strawberry.type
class Query:
    @strawberry.field
    def get_testing(
        self,
        info: strawberry.Info,
        id_: Annotated[UUID, strawberry.argument(name="id")],
    ) -> str | None: ...



import textwrap
from enum import Enum

import pydantic

import strawberry
from tests.experimental.pydantic.utils import needs_pydantic_v1


def test_basic_type_field_list():
    class UserModel(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1


def test_all_fields():
    class UserModel(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel, all_fields=True)
    class User:
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1


def test_auto_fields():
    class UserModel(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1


def test_basic_alias_type():
    class UserModel(pydantic.BaseModel):
        age_: int = pydantic.Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age_: strawberry.auto
        password: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


def test_basic_type_with_list():
    class UserModel(pydantic.BaseModel):
        age: int
        friend_names: list[str]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        friend_names: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, friend_names=["A", "B"])

    schema = strawberry.Schema(query=Query)

    query = "{ user { friendNames } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["friendNames"] == ["A", "B"]


def test_basic_type_with_nested_model():
    class Hobby(pydantic.BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(Hobby)
    class HobbyType:
        name: strawberry.auto

    class User(pydantic.BaseModel):
        hobby: Hobby

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        hobby: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            return UserType(hobby=HobbyType(name="Skii"))

    schema = strawberry.Schema(query=Query)

    query = "{ user { hobby { name } } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["hobby"]["name"] == "Skii"


def test_basic_type_with_list_of_nested_model():
    class Hobby(pydantic.BaseModel):
        name: str

    @strawberry.experimental.pydantic.type(Hobby)
    class HobbyType:
        name: strawberry.auto

    class User(pydantic.BaseModel):
        hobbies: list[Hobby]

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        hobbies: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            return UserType(
                hobbies=[
                    HobbyType(name="Skii"),
                    HobbyType(name="Cooking"),
                ]
            )

    schema = strawberry.Schema(query=Query)

    query = "{ user { hobbies { name } } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["hobbies"] == [
        {"name": "Skii"},
        {"name": "Cooking"},
    ]


def test_basic_type_with_extended_fields():
    class UserModel(pydantic.BaseModel):
        age: int

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        name: str
        age: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(name="Marco", age=100)

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      name: String!
      age: Int!
    }
    """
    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { name age } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["name"] == "Marco"
    assert result.data["user"]["age"] == 100


def test_type_with_custom_resolver():
    class UserModel(pydantic.BaseModel):
        age: int

    def get_age_in_months(root):
        return root.age * 12

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age_in_months: int = strawberry.field(resolver=get_age_in_months)
        age: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=20)

    schema = strawberry.Schema(query=Query)

    query = "{ user { age ageInMonths } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 20
    assert result.data["user"]["ageInMonths"] == 240


def test_basic_type_with_union():
    class BranchA(pydantic.BaseModel):
        field_a: str

    class BranchB(pydantic.BaseModel):
        field_b: int

    class User(pydantic.BaseModel):
        union_field: BranchA | BranchB

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType:
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType:
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        union_field: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            return UserType(union_field=BranchBType(field_b=10))

    schema = strawberry.Schema(query=Query)

    query = "{ user { unionField { ... on BranchBType { fieldB } } } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["unionField"]["fieldB"] == 10


def test_basic_type_with_union_pydantic_types():
    class BranchA(pydantic.BaseModel):
        field_a: str

    class BranchB(pydantic.BaseModel):
        field_b: int

    class User(pydantic.BaseModel):
        union_field: BranchA | BranchB

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType:
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType:
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        union_field: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            # note that BranchB is a pydantic type, not a strawberry type
            return UserType(union_field=BranchB(field_b=10))

    schema = strawberry.Schema(query=Query)

    query = "{ user { unionField { ... on BranchBType { fieldB } } } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["unionField"]["fieldB"] == 10


def test_basic_type_with_enum():
    @strawberry.enum
    class UserKind(Enum):
        user = 0
        admin = 1

    class User(pydantic.BaseModel):
        age: int
        kind: UserKind

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        kind: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            return UserType(age=10, kind=UserKind.admin)

    schema = strawberry.Schema(query=Query)

    query = "{ user { kind } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["kind"] == "admin"


def test_basic_type_with_interface():
    class Base(pydantic.BaseModel):
        base_field: str

    class BranchA(Base):
        field_a: str

    class BranchB(Base):
        field_b: int

    class User(pydantic.BaseModel):
        interface_field: Base

    @strawberry.experimental.pydantic.interface(Base)
    class BaseType:
        base_field: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType(BaseType):
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType(BaseType):
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        interface_field: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> UserType:
            return UserType(interface_field=BranchBType(base_field="abc", field_b=10))

    schema = strawberry.Schema(query=Query, types=[BranchAType, BranchBType])

    query = "{ user { interfaceField { baseField, ... on BranchBType { fieldB } } } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["interfaceField"]["baseField"] == "abc"
    assert result.data["user"]["interfaceField"]["fieldB"] == 10


def test_basic_type_with_optional_and_default():
    class UserModel(pydantic.BaseModel):
        age: int
        password: str | None = pydantic.Field(default="ABC")

    @strawberry.experimental.pydantic.type(UserModel, all_fields=True)
    class User:
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1)

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age password } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1
    assert result.data["user"]["password"] == "ABC"

    @strawberry.type
    class QueryNone:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password=None)

    schema = strawberry.Schema(query=QueryNone)

    query = "{ user { age password } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1
    assert result.data["user"]["password"] is None


@needs_pydantic_v1
def test_basic_type_with_constrained_list():
    class FriendList(pydantic.ConstrainedList):
        min_items = 1

    class UserModel(pydantic.BaseModel):
        age: int
        friend_names: FriendList[str]

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        friend_names: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, friend_names=["A", "B"])

    schema = strawberry.Schema(query=Query)

    query = "{ user { friendNames } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["friendNames"] == ["A", "B"]



import strawberry


def test_maybe_null_validation_rule_input_fields():
    """Test MaybeNullValidationRule validates input object fields correctly."""

    @strawberry.input
    class TestInput:
        strict_field: strawberry.Maybe[str]  # Should reject null
        flexible_field: strawberry.Maybe[str | None]  # Should allow null

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def test(self, input: TestInput) -> str:
            return "success"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    # Test 1: Valid values should work
    result = schema.execute_sync("""
        mutation {
            test(input: { strictField: "hello", flexibleField: "world" })
        }
    """)
    assert not result.errors
    assert result.data == {"test": "success"}

    # Test 2: Flexible field can be null
    result = schema.execute_sync("""
        mutation {
            test(input: { strictField: "hello", flexibleField: null })
        }
    """)
    assert not result.errors
    assert result.data == {"test": "success"}

    # Test 3: Strict field cannot be null
    result = schema.execute_sync("""
        mutation {
            test(input: { strictField: null, flexibleField: "world" })
        }
    """)
    assert result.errors
    assert len(result.errors) == 1
    error = result.errors[0]
    assert "Expected value of type 'str', found null" in str(error)
    assert "strictField" in str(error)
    assert "cannot be explicitly set to null" in str(error)
    assert "Use 'Maybe[str | None]'" in str(error)


def test_maybe_null_validation_rule_resolver_arguments():
    """Test MaybeNullValidationRule validates resolver arguments correctly."""

    @strawberry.type
    class Query:
        @strawberry.field
        def search(
            self,
            query: strawberry.Maybe[str] = None,  # Should reject null
            filter_by: strawberry.Maybe[str | None] = None,  # Should allow null
        ) -> str:
            return "success"

    schema = strawberry.Schema(query=Query)

    # Test 1: Valid values should work
    result = schema.execute_sync("""
        query {
            search(query: "hello", filterBy: "world")
        }
    """)
    assert not result.errors
    assert result.data == {"search": "success"}

    # Test 2: Flexible argument can be null
    result = schema.execute_sync("""
        query {
            search(query: "hello", filterBy: null)
        }
    """)
    assert not result.errors
    assert result.data == {"search": "success"}

    # Test 3: Strict argument cannot be null
    result = schema.execute_sync("""
        query {
            search(query: null, filterBy: "world")
        }
    """)
    assert result.errors
    assert len(result.errors) == 1
    error = result.errors[0]
    assert "Expected value of type 'str', found null" in str(error)
    assert "query" in str(error)
    assert "cannot be explicitly set to null" in str(error)
    assert "Use 'Maybe[str | None]'" in str(error)


def test_maybe_null_validation_rule_multiple_errors():
    """Test that multiple null violations are all reported."""

    @strawberry.input
    class TestInput:
        field1: strawberry.Maybe[str]
        field2: strawberry.Maybe[int]
        field3: strawberry.Maybe[str | None]  # This one allows null

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def test(self, input: TestInput) -> str:
            return "success"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    # Test with multiple nulls - should get multiple errors
    result = schema.execute_sync("""
        mutation {
            test(input: { field1: null, field2: null, field3: null })
        }
    """)
    assert result.errors
    assert len(result.errors) == 2  # field1 and field2 should fail, field3 should pass

    error_messages = [str(error) for error in result.errors]
    assert any("field1" in msg for msg in error_messages)
    assert any("field2" in msg for msg in error_messages)
    # field3 should NOT generate an error because it allows null


def test_maybe_null_validation_rule_nested_input():
    """Test validation works with nested input objects."""

    @strawberry.input
    class NestedInput:
        value: strawberry.Maybe[str]

    @strawberry.input
    class TestInput:
        nested: NestedInput

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def test(self, input: TestInput) -> str:
            return "success"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    # Test with null in nested input
    result = schema.execute_sync("""
        mutation {
            test(input: { nested: { value: null } })
        }
    """)
    assert result.errors
    assert len(result.errors) == 1
    error = result.errors[0]
    assert "Expected value of type 'str', found null" in str(error)
    assert "value" in str(error)


def test_maybe_null_validation_rule_different_types():
    """Test validation works with different field types."""

    @strawberry.input
    class TestInput:
        string_field: strawberry.Maybe[str]
        int_field: strawberry.Maybe[int]
        bool_field: strawberry.Maybe[bool]
        list_field: strawberry.Maybe[list[str]]

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def test(self, input: TestInput) -> str:
            return "success"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    # Test each field type with null
    test_cases = [
        ("stringField", "str"),
        ("intField", "int"),
        ("boolField", "bool"),
        ("listField", "list[str]"),
    ]

    for field_name, type_name in test_cases:
        result = schema.execute_sync(f"""
            mutation {{
                test(input: {{ {field_name}: null }})
            }}
        """)
        assert result.errors
        assert len(result.errors) == 1
        error = result.errors[0]
        assert f"Expected value of type '{type_name}', found null" in str(error)


def test_maybe_null_validation_rule_custom_graphql_names():
    """Test validation works with custom GraphQL field names."""

    @strawberry.input
    class TestInput:
        internal_name: strawberry.Maybe[str] = strawberry.field(name="customName")

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def test(self, input: TestInput) -> str:
            return "success"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    # Test with custom GraphQL name
    result = schema.execute_sync("""
        mutation {
            test(input: { customName: null })
        }
    """)
    assert result.errors
    assert len(result.errors) == 1
    error = result.errors[0]
    assert "customName" in str(error)


# TODO: Add test for auto_camel_case=False configuration
# This requires accessing the schema configuration from the validation rule context,
# which needs further investigation of the GraphQL validation context API.



import pytest
from graphql import (
    GraphQLError,
    get_introspection_query,
    parse,
    specified_rules,
    validate,
)

import strawberry
from strawberry.extensions import QueryDepthLimiter
from strawberry.extensions.query_depth_limiter import (
    IgnoreContext,
    ShouldIgnoreType,
    create_validator,
)


@strawberry.interface
class Pet:
    name: str
    owner: "Human"


@strawberry.type
class Cat(Pet):
    pass


@strawberry.type
class Dog(Pet):
    pass


@strawberry.type
class Address:
    street: str
    number: int
    city: str
    country: str


@strawberry.type
class Human:
    name: str
    email: str
    address: Address
    pets: list[Pet]


@strawberry.input
class Biography:
    name: str
    owner_name: str


@strawberry.type
class Query:
    @strawberry.field
    def user(
        self,
        name: str | None,
        id: int | None,
        age: float | None,
        is_cool: bool | None,
    ) -> Human:
        pass

    @strawberry.field
    def users(self, names: list[str] | None) -> list[Human]:
        pass

    @strawberry.field
    def cat(bio: Biography) -> Cat:
        pass

    version: str
    user1: Human
    user2: Human
    user3: Human


schema = strawberry.Schema(Query)


def run_query(
    query: str, max_depth: int, should_ignore: ShouldIgnoreType = None
) -> tuple[list[GraphQLError], dict[str, int] | None]:
    document = parse(query)

    result = None

    def callback(query_depths):
        nonlocal result
        result = query_depths

    validation_rule = create_validator(max_depth, should_ignore, callback)

    errors = validate(
        schema._schema,
        document,
        rules=(*specified_rules, validation_rule),
    )

    return errors, result


def test_should_count_depth_without_fragment():
    query = """
    query read0 {
      version
    }
    query read1 {
      version
      user {
        name
      }
    }
    query read2 {
      matt: user(name: "matt") {
        email
      }
      andy: user(name: "andy") {
        email
        address {
          city
        }
      }
    }
    query read3 {
      matt: user(name: "matt") {
        email
      }
      andy: user(name: "andy") {
        email
        address {
          city
        }
        pets {
          name
          owner {
            name
          }
        }
      }
    }
    """

    expected = {"read0": 0, "read1": 1, "read2": 2, "read3": 3}

    errors, result = run_query(query, 10)
    assert not errors
    assert result == expected


def test_should_count_with_fragments():
    query = """
    query read0 {
      ... on Query {
        version
      }
    }
    query read1 {
      version
      user {
        ... on Human {
          name
        }
      }
    }
    fragment humanInfo on Human {
      email
    }
    fragment petInfo on Pet {
      name
      owner {
        name
      }
    }
    query read2 {
      matt: user(name: "matt") {
        ...humanInfo
      }
      andy: user(name: "andy") {
        ...humanInfo
        address {
          city
        }
      }
    }
    query read3 {
      matt: user(name: "matt") {
        ...humanInfo
      }
      andy: user(name: "andy") {
        ... on Human {
          email
        }
        address {
          city
        }
        pets {
          ...petInfo
        }
      }
    }
  """

    expected = {"read0": 0, "read1": 1, "read2": 2, "read3": 3}

    errors, result = run_query(query, 10)
    assert not errors
    assert result == expected


def test_should_ignore_the_introspection_query():
    errors, result = run_query(get_introspection_query(), 10)
    assert not errors
    assert result == {"IntrospectionQuery": 0}


def test_should_catch_query_thats_too_deep():
    query = """{
    user {
      pets {
        owner {
          pets {
            owner {
              pets {
                name
              }
            }
          }
        }
      }
    }
    }
    """
    errors, _result = run_query(query, 4)

    assert len(errors) == 1
    assert errors[0].message == "'anonymous' exceeds maximum operation depth of 4"


def test_should_raise_invalid_ignore():
    with pytest.raises(
        TypeError,
        match=r"The `should_ignore` argument to `QueryDepthLimiter` must be a callable.",
    ):
        strawberry.Schema(
            Query, extensions=[QueryDepthLimiter(max_depth=10, should_ignore=True)]
        )


def test_should_ignore_field_by_name():
    query = """
    query read1 {
      user { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_name in ("user1", "user2", "user3")

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 2, "read2": 0}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_str_argument():
    query = """
    query read1 {
      user(name:"matt") { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("name") == "matt"

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_int_argument():
    query = """
    query read1 {
      user(id:1) { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("id") == 1

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_float_argument():
    query = """
    query read1 {
      user(age:10.5) { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("age") == 10.5

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_bool_argument():
    query = """
    query read1 {
      user(isCool:false) { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("isCool") is False

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_name_and_str_argument():
    query = """
    query read1 {
      user(name:"matt") { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("name") == "matt"

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_list_argument():
    query = """
    query read1 {
      users(names:["matt","andy"]) { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return "matt" in ignore.field_args.get("names", [])

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_ignore_field_by_object_argument():
    query = """
    query read1 {
      cat(bio:{
        name:"Momo",
        ownerName:"Tommy"
      }) { name }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return ignore.field_args.get("bio", {}).get("name") == "Momo"

    errors, result = run_query(query, 10, should_ignore=should_ignore)

    expected = {"read1": 0, "read2": 2}
    assert not errors
    assert result == expected


def test_should_work_as_extension():
    query = """{
    user {
      pets {
        owner {
          pets {
            owner {
              pets {
                name
              }
            }
          }
        }
      }
    }
    }
    """

    def should_ignore(ignore: IgnoreContext) -> bool:
        return False

    schema = strawberry.Schema(
        Query, extensions=[QueryDepthLimiter(max_depth=4, should_ignore=should_ignore)]
    )

    result = schema.execute_sync(query)

    assert len(result.errors) == 1
    assert (
        result.errors[0].message == "'anonymous' exceeds maximum operation depth of 4"
    )



from enum import Enum, auto
from typing import Annotated, cast

from graphql import (
    DirectiveLocation,
    GraphQLEnumType,
    GraphQLInputType,
    GraphQLObjectType,
    GraphQLSchema,
)

import strawberry
from strawberry.directive import DirectiveValue
from strawberry.scalars import JSON
from strawberry.schema.schema_converter import GraphQLCoreConverter
from strawberry.schema_directive import Location
from strawberry.types.base import get_object_definition

DEFINITION_BACKREF = GraphQLCoreConverter.DEFINITION_BACKREF


def test_extensions_schema_directive():
    @strawberry.schema_directive(locations=[Location.OBJECT, Location.INPUT_OBJECT])
    class SchemaDirective:
        name: str

    @strawberry.type(directives=[SchemaDirective(name="Query")])
    class Query:
        hello: str

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema

    # Schema
    assert graphql_schema.extensions[DEFINITION_BACKREF] is schema

    # TODO: Apparently I stumbled on a bug:
    #        SchemaDirective are used on schema.__str__(),
    #        but aren't added to graphql_schema.directives
    # maybe graphql_schema_directive = graphql_schema.get_directive("schemaDirective")

    directives = get_object_definition(Query, strict=True).directives
    assert directives is not None
    graphql_schema_directive = schema.schema_converter.from_schema_directive(
        directives[0]
    )
    assert (
        graphql_schema_directive.extensions[DEFINITION_BACKREF]
        is SchemaDirective.__strawberry_directive__
    )


def test_directive():
    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def uppercase(value: DirectiveValue[str], foo: str):  # pragma: no cover
        return value.upper()

    @strawberry.type()
    class Query:
        hello: str

    schema = strawberry.Schema(query=Query, directives=[uppercase])
    graphql_schema: GraphQLSchema = schema._schema

    graphql_directive = graphql_schema.get_directive("uppercase")
    assert graphql_directive.extensions[DEFINITION_BACKREF] is uppercase
    assert (
        graphql_directive.args["foo"].extensions[DEFINITION_BACKREF]
        is uppercase.arguments[0]
    )


def test_enum():
    @strawberry.enum
    class ThingType(Enum):
        JSON = auto()
        STR = auto()

    @strawberry.type()
    class Query:
        hello: ThingType

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema

    graphql_thing_type = cast("GraphQLEnumType", graphql_schema.get_type("ThingType"))
    assert (
        graphql_thing_type.extensions[DEFINITION_BACKREF]
        is ThingType.__strawberry_definition__
    )
    assert (
        graphql_thing_type.values["JSON"].extensions[DEFINITION_BACKREF]
        is ThingType.__strawberry_definition__.values[0]
    )
    assert (
        graphql_thing_type.values["STR"].extensions[DEFINITION_BACKREF]
        is ThingType.__strawberry_definition__.values[1]
    )


def test_scalar():
    from strawberry.schema.types.scalar import DEFAULT_SCALAR_REGISTRY

    @strawberry.type()
    class Query:
        hello: JSON
        hi: str

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema

    assert (
        graphql_schema.get_type("JSON").extensions[DEFINITION_BACKREF]
        is DEFAULT_SCALAR_REGISTRY[JSON]
    )


def test_interface():
    @strawberry.interface
    class Thing:
        name: str

    @strawberry.type()
    class Query:
        hello: Thing

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema

    assert (
        graphql_schema.get_type("Thing").extensions[DEFINITION_BACKREF]
        is Thing.__strawberry_definition__
    )


def test_union():
    @strawberry.type
    class JsonThing:
        value: JSON

    @strawberry.type
    class StrThing:
        value: str

    SomeThing = Annotated[JsonThing | StrThing, strawberry.union("SomeThing")]

    @strawberry.type()
    class Query:
        hello: SomeThing

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema
    graphql_type = graphql_schema.get_type("SomeThing")

    assert graphql_type.extensions[DEFINITION_BACKREF].graphql_name == "SomeThing"
    assert graphql_type.extensions[DEFINITION_BACKREF].description is None


def test_object_types():
    @strawberry.input
    class Input:
        name: str

    @strawberry.type()
    class Query:
        @strawberry.field
        def hello(self, input: Input) -> str: ...

    schema = strawberry.Schema(query=Query)
    graphql_schema: GraphQLSchema = schema._schema

    assert (
        graphql_schema.get_type("Input").extensions[DEFINITION_BACKREF]
        is Input.__strawberry_definition__
    )
    assert (
        graphql_schema.get_type("Query").extensions[DEFINITION_BACKREF]
        is Query.__strawberry_definition__
    )

    graphql_query = cast("GraphQLObjectType", graphql_schema.get_type("Query"))
    assert graphql_query.fields["hello"].extensions[
        DEFINITION_BACKREF
    ] is Query.__strawberry_definition__.get_field("hello")
    assert (
        graphql_query.fields["hello"].args["input"].extensions[DEFINITION_BACKREF]
        is Query.__strawberry_definition__.get_field("hello").arguments[0]
    )

    graphql_input = cast("GraphQLInputType", graphql_schema.get_type("Input"))
    assert graphql_input.fields["name"].extensions[
        DEFINITION_BACKREF
    ] is Input.__strawberry_definition__.get_field("name")



import strawberry
from strawberry.extensions import SchemaExtension


@strawberry.type
class Query:
    @strawberry.field
    def ping(self) -> str:
        return "pong"


def test_execution_context_operation_name_and_type():
    operation_name = None
    operation_type = None

    class MyExtension(SchemaExtension):
        def on_operation(self):
            yield
            nonlocal operation_name
            nonlocal operation_type

            execution_context = self.execution_context

            operation_name = execution_context.operation_name
            operation_type = execution_context.operation_type.value

    schema = strawberry.Schema(Query, extensions=[MyExtension])

    result = schema.execute_sync("{ ping }")
    assert not result.errors

    assert operation_name is None
    assert operation_type == "query"

    # Try again with an operation_name
    result = schema.execute_sync("query MyOperation { ping }")
    assert not result.errors

    assert operation_name == "MyOperation"
    assert operation_type == "query"

    # Try again with an operation_name override
    result = schema.execute_sync(
        """
        query MyOperation { ping }
        query MyOperation2 { ping }
        """,
        operation_name="MyOperation2",
    )
    assert not result.errors

    assert operation_name == "MyOperation2"
    assert operation_type == "query"


def test_execution_context_operation_type_mutation():
    operation_name = None
    operation_type = None

    class MyExtension(SchemaExtension):
        def on_operation(self):
            yield
            nonlocal operation_name
            nonlocal operation_type

            execution_context = self.execution_context

            operation_name = execution_context.operation_name
            operation_type = execution_context.operation_type.value

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def my_mutation(self) -> str:
            return "hi"

    schema = strawberry.Schema(Query, mutation=Mutation, extensions=[MyExtension])

    result = schema.execute_sync("mutation { myMutation }")
    assert not result.errors

    assert operation_name is None
    assert operation_type == "mutation"

    # Try again with an operation_name
    result = schema.execute_sync("mutation MyMutation { myMutation }")
    assert not result.errors

    assert operation_name == "MyMutation"
    assert operation_type == "mutation"

    # Try again with an operation_name override
    result = schema.execute_sync(
        """
        mutation MyMutation { myMutation }
        mutation MyMutation2 { myMutation }
        """,
        operation_name="MyMutation2",
    )
    assert not result.errors

    assert operation_name == "MyMutation2"
    assert operation_type == "mutation"


def test_execution_context_operation_name_and_type_with_fragments():
    operation_name = None
    operation_type = None

    class MyExtension(SchemaExtension):
        def on_operation(self):
            yield
            nonlocal operation_name
            nonlocal operation_type

            execution_context = self.execution_context

            operation_name = execution_context.operation_name
            operation_type = execution_context.operation_type.value

    schema = strawberry.Schema(Query, extensions=[MyExtension])

    result = schema.execute_sync(
        """
        fragment MyFragment on Query {
            ping
        }

        query MyOperation {
            ping
            ...MyFragment
        }
        """
    )
    assert not result.errors

    assert operation_name == "MyOperation"
    assert operation_type == "query"


def test_error_when_accessing_operation_type_before_parsing():
    class MyExtension(SchemaExtension):
        def on_operation(self):
            execution_context = self.execution_context

            # This should raise a RuntimeError
            execution_context.operation_type

    schema = strawberry.Schema(Query, extensions=[MyExtension])

    result = schema.execute_sync("mutation { myMutation }")
    assert len(result.errors) == 1
    assert isinstance(result.errors[0].original_error, RuntimeError)
    assert result.errors[0].message == "No GraphQL document available"


def test_error_when_accessing_operation_type_with_invalid_operation_name():
    class MyExtension(SchemaExtension):
        def on_parse(self):
            yield
            execution_context = self.execution_context

            # This should raise a RuntimeError
            execution_context.operation_type

    schema = strawberry.Schema(Query, extensions=[MyExtension])

    result = schema.execute_sync("query { ping }", operation_name="MyQuery")
    assert len(result.errors) == 1
    assert isinstance(result.errors[0].original_error, RuntimeError)
    assert result.errors[0].message == "Can't get GraphQL operation type"



from __future__ import annotations
import strawberry

@strawberry.type(description="A connection to a list of items.")
class FilmCharactersConnection:
    page_info: PageInfo = strawberry.field(description="Information to aid in pagination.")
    edges: list[FilmCharactersEdge | None] | None = strawberry.field(description="A list of edges.")
    total_count: int | None = strawberry.field(description="""
A count of the total number of objects in this connection, ignoring pagination.
This allows a client to fetch the first five objects by passing "5" as the
argument to "first", then fetch the total count so it could display "5 of 83",
for example.
""")
    characters: list[Person | None] | None = strawberry.field(description="""
A list of all of the objects returned in the connection. This is a convenience
field provided for quickly exploring the API; rather than querying for
"{ edges { node } }" when no edge data is needed, this field can be be used
instead. Note that when clients like Relay need to fetch the "cursor" field on
the edge to enable efficient pagination, this shortcut cannot be used, and the
full "{ edges { node } }" version should be used instead.
""")



from __future__ import annotations

import textwrap

import strawberry
from strawberry.schema.config import StrawberryConfig
from tests.conftest import skip_if_gql_32

pytestmark = skip_if_gql_32("GraphQL 3.3.0 is required for incremental execution")


@strawberry.type
class Query:
    hello: str


def test_does_not_print_defer_and_stream_directives_when_experimental_execution_is_disabled():
    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(enable_experimental_incremental_execution=False),
    )

    expected_type = """
    type Query {
      hello: String!
    }
    """

    assert str(schema) == textwrap.dedent(expected_type).strip()


def test_prints_defer_and_stream_directives_when_experimental_execution_is_enabled():
    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(enable_experimental_incremental_execution=True),
    )

    expected_type = """
    directive @defer(if: Boolean, label: String) on FRAGMENT_SPREAD | INLINE_FRAGMENT

    directive @stream(if: Boolean, label: String, initialCount: Int = 0) on FIELD

    type Query {
      hello: String!
    }
    """

    assert str(schema) == textwrap.dedent(expected_type).strip()



from typing import Any

import pytest

import strawberry
from strawberry.schema_directives import OneOf


@strawberry.input(one_of=True)
class ExampleInputTagged:
    a: strawberry.Maybe[str | None]
    b: strawberry.Maybe[int | None]


@strawberry.type
class ExampleResult:
    a: str | None
    b: int | None


@strawberry.type
class Query:
    @strawberry.field
    def test(self, input: ExampleInputTagged) -> ExampleResult:
        if input.a:
            return ExampleResult(a=input.a.value, b=None)
        if input.b:
            return ExampleResult(a=None, b=input.b.value)
        return ExampleResult(a=None, b=None)


schema = strawberry.Schema(query=Query)


@pytest.mark.parametrize(
    ("default_value", "variables"),
    [
        ("{a: null, b: null}", {}),
        ('{ a: "abc", b: 123 }', {}),
        ("{a: null, b: 123}", {}),
        ("{}", {}),
    ],
)
def test_must_specify_at_least_one_key_default(
    default_value: str, variables: dict[str, Any]
):
    query = f"""
        query ($input: ExampleInputTagged! = {default_value}) {{
          test(input: $input) {{
            a
            b
          }}
        }}
    """

    result = schema.execute_sync(query, variable_values=variables)

    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "OneOf Input Object 'ExampleInputTagged' must specify exactly one key."
    )


@pytest.mark.parametrize(
    ("value", "variables"),
    [
        ("{a: null, b: null}", {}),
        ('{ a: "abc", b: 123 }', {}),
        ("{a: null, b: 123}", {}),
        ("{}", {}),
        ("{ a: $a, b: 123 }", {"a": "abc"}),
        ("{ a: $a, b: 123 }", {}),
        ("{ a: $a, b: $b }", {"a": "abc"}),
        ("$input", {"input": {"a": "abc", "b": 123}}),
        ("$input", {"input": {"a": "abc", "b": None}}),
        ("$input", {"input": {}}),
        ('{ a: "abc", b: null }', {}),
    ],
)
def test_must_specify_at_least_one_key_literal(value: str, variables: dict[str, Any]):
    variables_definitions = []

    if "$a" in value:
        variables_definitions.append("$a: String")

    if "$b" in value:
        variables_definitions.append("$b: Int")

    if "$input" in value:
        variables_definitions.append("$input: ExampleInputTagged!")

    variables_definition_str = (
        f"({', '.join(variables_definitions)})" if variables_definitions else ""
    )

    query = f"""
        query {variables_definition_str} {{
          test(input: {value}) {{
            a
            b
          }}
        }}
    """

    result = schema.execute_sync(query, variable_values=variables)

    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "OneOf Input Object 'ExampleInputTagged' must specify exactly one key."
    )


def test_value_must_be_non_null_input():
    query = """
        query ($input: ExampleInputTagged!) {
          test(input: $input) {
            a
            b
          }
        }
    """

    result = schema.execute_sync(query, variable_values={"input": {"a": None}})

    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Value for member field 'a' must be non-null"


def test_value_must_be_non_null_literal():
    query = """
        query {
          test(input: { a: null }) {
            a
            b
          }
        }
    """

    result = schema.execute_sync(query, variable_values={"input": {"a": None}})

    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Field 'ExampleInputTagged.a' must be non-null."


def test_value_must_be_non_null_variable():
    query = """
        query ($b: Int) {
            test(input: { b: $b }) {
                b
            }
        }
    """

    result = schema.execute_sync(query, variable_values={})

    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Variable 'b' must be non-nullable to be used for OneOf Input Object 'ExampleInputTagged'."
    )


@pytest.mark.parametrize(
    ("value", "variables", "expected"),
    [
        ("{ b: $b }", {"b": 123}, {"b": 123}),
        ("$input", {"input": {"b": 123}}, {"b": 123}),
        ('{ a: "abc" }', {}, {"a": "abc"}),
        ("$input", {"input": {"a": "abc"}}, {"a": "abc"}),
    ],
)
def test_works(value: str, variables: dict[str, Any], expected: dict[str, Any]):
    variables_definitions = []

    if "$b" in value:
        variables_definitions.append("$b: Int!")

    if "$input" in value:
        variables_definitions.append("$input: ExampleInputTagged!")

    variables_definition_str = (
        f"({', '.join(variables_definitions)})" if variables_definitions else ""
    )

    field = next(iter(expected.keys()))

    query = f"""
        query {variables_definition_str} {{
          test(input: {value}) {{
            {field}
          }}
        }}
    """

    result = schema.execute_sync(query, variable_values=variables)

    assert not result.errors
    assert result.data["test"] == expected


def test_works_with_camelcasing():
    global ExampleWithLongerNames, Result

    @strawberry.input(directives=[OneOf()])
    class ExampleWithLongerNames:
        a_field: strawberry.Maybe[str | None]
        b_field: strawberry.Maybe[int | None]

    @strawberry.type
    class Result:
        a_field: str | None
        b_field: int | None

    @strawberry.type
    class Query:
        @strawberry.field
        def test(self, input: ExampleWithLongerNames) -> Result:
            return Result(  # noqa: F821
                a_field=input.a_field.value if input.a_field else None,
                b_field=input.b_field.value if input.b_field else None,
            )

    schema = strawberry.Schema(query=Query)

    query = """
        query ($input: ExampleWithLongerNames!) {
          test(input: $input) {
            aField
            bField
          }
        }
    """

    result = schema.execute_sync(query, variable_values={"input": {"aField": "abc"}})

    assert not result.errors
    assert result.data["test"] == {"aField": "abc", "bField": None}

    del ExampleWithLongerNames, Result


def test_introspection():
    query = """
        query {
          __type(name: "ExampleInputTagged") {
            name
            isOneOf
          }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data == {"__type": {"name": "ExampleInputTagged", "isOneOf": True}}


def test_introspection_builtin():
    query = """
        query {
          __type(name: "String") {
            name
            isOneOf
          }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data == {"__type": {"name": "String", "isOneOf": False}}



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy]

MYPY_PLUGINS = ["pydantic.mypy", "strawberry.ext.mypy_plugin"]

CODE = """
import pydantic
import strawberry

class UserModel(pydantic.BaseModel):
    age: int
    name: str

@strawberry.experimental.pydantic.type(model=UserModel)
class User:
    age: strawberry.auto
    name: strawberry.auto

user = User(age=1, name="abc")
reveal_type(user)
reveal_type(user.to_pydantic())
reveal_type(User.from_pydantic(UserModel(age=1, name="abc")))
"""


def test_pydantic_type():
    results = typecheck(CODE, mypy_plugins=MYPY_PLUGINS)

    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "mypy_test.User"',
                line=15,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "mypy_test.UserModel"',
                line=16,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "mypy_test.User"',
                line=17,
                column=13,
            ),
        ]
    )

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "user" is "StrawberryTypeFromPydantic[UserModel]"',
                line=15,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "user.to_pydantic()" is "UserModel"',
                line=16,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "User.from_pydantic(UserModel(age=1, name="abc"))" is "StrawberryTypeFromPydantic[UserModel]"',
                line=17,
                column=13,
            ),
        ]
    )

    # TY doesn't properly support this yet
    # assert results.ty == snapshot(...)  # noqa: ERA001



import textwrap

import pydantic

import strawberry
from strawberry.printer import print_schema
from tests.conftest import skip_if_gql_32
from tests.experimental.pydantic.utils import needs_pydantic_v2


def test_field_type_default():
    class User(pydantic.BaseModel):
        name: str = "James"
        nickname: str | None = "Jim"

    @strawberry.experimental.pydantic.type(User, all_fields=True)
    class PydanticUser: ...

    @strawberry.type
    class StrawberryUser:
        name: str = "James"

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self) -> PydanticUser:
            return PydanticUser()

        @strawberry.field
        def b(self) -> StrawberryUser:
            return StrawberryUser()

    schema = strawberry.Schema(Query)

    # name should be required in both the PydanticUser and StrawberryUser
    expected = """
    type PydanticUser {
      name: String!
      nickname: String
    }

    type Query {
      a: PydanticUser!
      b: StrawberryUser!
    }

    type StrawberryUser {
      name: String!
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_pydantic_type_default_none():
    class UserPydantic(pydantic.BaseModel):
        name: str | None = None

    @strawberry.experimental.pydantic.type(UserPydantic, all_fields=True)
    class User: ...

    @strawberry.type
    class Query:
        a: User = strawberry.field()

    schema = strawberry.Schema(Query)

    expected = """
    type Query {
      a: User!
    }

    type User {
      name: String
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_pydantic_type_no_default_but_optional():
    class UserPydantic(pydantic.BaseModel):
        # pydantic automatically adds a default of None for Optional fields
        name: str | None

    @strawberry.experimental.pydantic.type(UserPydantic, all_fields=True)
    class User: ...

    @strawberry.type
    class Query:
        a: User = strawberry.field()

    schema = strawberry.Schema(Query)

    expected = """
    type Query {
      a: User!
    }

    type User {
      name: String
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_input_type_default():
    class User(pydantic.BaseModel):
        name: str = "James"

    @strawberry.experimental.pydantic.type(User, all_fields=True, is_input=True)
    class PydanticUser: ...

    @strawberry.type(is_input=True)
    class StrawberryUser:
        name: str = "James"

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self, user: PydanticUser) -> str:
            return user.name

        @strawberry.field
        def b(self, user: StrawberryUser) -> str:
            return user.name

    schema = strawberry.Schema(Query)

    # name should be required in both the PydanticUser and StrawberryUser
    expected = """
    input PydanticUser {
      name: String! = "James"
    }

    type Query {
      a(user: PydanticUser!): String!
      b(user: StrawberryUser!): String!
    }

    input StrawberryUser {
      name: String! = "James"
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()


@needs_pydantic_v2
def test_v2_explicit_default():
    class User(pydantic.BaseModel):
        name: str | None

    @strawberry.experimental.pydantic.type(User, all_fields=True)
    class PydanticUser: ...

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self) -> PydanticUser:
            raise NotImplementedError

    schema = strawberry.Schema(Query)

    # name should have no default
    expected = """
    type PydanticUser {
      name: String
    }

    type Query {
      a: PydanticUser!
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()


@skip_if_gql_32("formatting is different in gql 3.2")
def test_v2_input_with_nonscalar_default():
    class NonScalarType(pydantic.BaseModel):
        id: int = 10
        nullable_field: int | None = None

    class Owning(pydantic.BaseModel):
        non_scalar_type: NonScalarType = NonScalarType()
        id: int = 10

    @strawberry.experimental.pydantic.type(
        model=NonScalarType, all_fields=True, is_input=True
    )
    class NonScalarTypeInput: ...

    @strawberry.experimental.pydantic.type(model=Owning, all_fields=True, is_input=True)
    class OwningInput: ...

    @strawberry.type
    class ExampleOutput:
        owning_id: int
        non_scalar_id: int
        non_scalar_nullable_field: int | None

    @strawberry.type
    class Query:
        @strawberry.field()
        def test(self, x: OwningInput) -> ExampleOutput:
            return ExampleOutput(
                owning_id=x.id,
                non_scalar_id=x.non_scalar_type.id,
                non_scalar_nullable_field=x.non_scalar_type.nullable_field,
            )

    schema = strawberry.Schema(Query)

    expected = """
    type ExampleOutput {
      owningId: Int!
      nonScalarId: Int!
      nonScalarNullableField: Int
    }

    input NonScalarTypeInput {
      id: Int! = 10
      nullableField: Int = null
    }

    input OwningInput {
      nonScalarType: NonScalarTypeInput! = { id: 10 }
      id: Int! = 10
    }

    type Query {
      test(x: OwningInput!): ExampleOutput!
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected).strip()

    query = """
    query($input_data: OwningInput!)
    {
        test(x: $input_data) {
            owningId nonScalarId nonScalarNullableField
        }
    }
    """
    result = schema.execute_sync(
        query, variable_values={"input_data": {"nonScalarType": {}}}
    )

    assert not result.errors
    expected_result = {
        "owningId": 10,
        "nonScalarId": 10,
        "nonScalarNullableField": None,
    }
    assert result.data["test"] == expected_result



from pydantic import BaseModel, Field

import strawberry
from strawberry.experimental.pydantic import input as pydantic_input
from strawberry.extensions import PydanticErrorExtension


class UserModel(BaseModel):
    age: int = Field(gt=18)
    score: int = Field(gt=50)


@pydantic_input(model=UserModel)
class UserInput:
    age: int
    score: int


@strawberry.type
class Query:
    hello: str = "hi"


@strawberry.type
class Mutation:
    @strawberry.mutation
    def create_user(self, data: UserInput) -> bool:
        UserModel(**data.__dict__)
        return True


schema = strawberry.Schema(
    query=Query,
    mutation=Mutation,
    extensions=[PydanticErrorExtension],
)


def test_single_validation_error():
    result = schema.execute_sync(
        """
        mutation {
            createUser(data: { age: 10, score: 60 })
        }
        """
    )

    assert result.errors is not None
    extensions = result.errors[0].extensions or {}
    assert "validation_errors" in extensions

    errors = extensions["validation_errors"]
    assert len(errors) == 1
    assert errors[0]["field"] == "age"


def test_multiple_validation_errors():
    result = schema.execute_sync(
        """
        mutation {
            createUser(data: { age: 10, score: 40 })
        }
        """
    )

    assert result.errors is not None
    extensions = result.errors[0].extensions or {}
    assert "validation_errors" in extensions

    errors = extensions["validation_errors"]
    fields = [e["field"] for e in errors]

    assert "age" in fields
    assert "score" in fields
    assert len(errors) == 2



from typing import Any

from litestar import Litestar, Request
from litestar.di import Provide

from strawberry.litestar import make_graphql_controller
from tests.views.schema import schema


def custom_context_dependency() -> str:
    return "Hi!"


async def get_root_value(request: Request = None):
    return request


async def get_context(app_dependency: str, request: Request = None):
    return {"custom_value": app_dependency, "request": request}


def create_app(schema=schema, **kwargs: Any):
    GraphQLController = make_graphql_controller(
        schema,
        path="/graphql",
        context_getter=get_context,
        root_value_getter=get_root_value,
        **kwargs,
    )

    return Litestar(
        route_handlers=[GraphQLController],
        dependencies={
            "app_dependency": Provide(custom_context_dependency, sync_to_thread=True)
        },
    )



import pytest

import strawberry
from strawberry.exceptions import (
    FieldWithResolverAndDefaultFactoryError,
    FieldWithResolverAndDefaultValueError,
    InvalidDefaultFactoryError,
)
from strawberry.types.field import StrawberryField


def test_field_with_default():
    @strawberry.type
    class Query:
        the_field: int = strawberry.field(default=3)

    instance = Query()
    assert instance.the_field == 3


def test_field_with_resolver_and_default():
    with pytest.raises(FieldWithResolverAndDefaultValueError):

        @strawberry.type
        class Query:
            @strawberry.field(default="potato")
            def fruit(self) -> str:
                return "tomato"


def test_field_with_default_factory():
    @strawberry.type
    class Query:
        the_int: int = strawberry.field(default_factory=lambda: 3)

    instance = Query()
    [int_field] = Query.__strawberry_definition__.fields

    assert instance.the_int == 3
    assert int_field.default_value == 3


def test_field_default_extensions_value_set():
    field = StrawberryField(python_name="test", default="test")
    assert field.extensions == []


def test_field_default_factory_executed_each_time():
    @strawberry.type
    class Query:
        the_list: list[str] = strawberry.field(default_factory=list)

    assert Query().the_list == Query().the_list
    assert Query().the_list is not Query().the_list


def test_field_with_separate_resolver_default():
    def fruit_resolver() -> str:  # pragma: no cover
        return "banana"

    with pytest.raises(FieldWithResolverAndDefaultValueError):

        @strawberry.type
        class Query:
            weapon: str = strawberry.field(
                default="strawberry", resolver=fruit_resolver
            )


def test_field_with_resolver_and_default_factory():
    with pytest.raises(FieldWithResolverAndDefaultFactoryError):

        @strawberry.type
        class Query:
            @strawberry.field(default_factory=lambda: "steel")
            def metal(self) -> str:
                return "iron"


def test_invalid_default_factory():
    with pytest.raises(InvalidDefaultFactoryError):
        strawberry.field(default_factory=round)
